/* appsocket.c
 *
 * Single source module for the Tektronix "appsocket" tool.
 *
 * Copyright (C) 1995,96 Tektronix, Inc. All rights reserved, under the terms
 * and conditions expressed below. This source code is placed in the public
 * domain for use without charge by any individual or organization. 
 *
 * 10/9/95 Kevin A. Draz v1.0	  - Initial release.
 * 11/14/96 KAD v 1.1			  - fixed EOF conditions bug (duplexing)
 *                                added stream to -f option ("-f -")
 *                                and progress reporting. Added -x option.
 * 11/18/96 KAD v1.1.1			  - added code in connection phase
 *                                to correct idle/not authorized error.
 * 11/20/96 KAD v1.1.2			  - more correction to idle printer/connection
 * 	  								    refused logic.
 *	11/21/96 KAD v1.1.3          - final (?) fixes to the connect()/idle
 *											 logic and added a fflush to stdout to
 *											 try to sync up return msg. from printers
 * 11/21/96 KAD v1.2				  - added the timer to the status, fixed
 *											 progress indicator so not so noisy
 * 11/22/96 KAD v1.2.1			  - added ioctl flush to net send stream
 *											 to fix an occasional last packet lost
 *
 * "appsocket" is a Unix command line tool which can be used to
 * send files to, and request status from, a Tektronix TCP/IP
 * enabled printer using the AppSocket protocol. Interactive (executive)
 * terminal mode to a printer is also supported.
 *
 * This source code is also provided as an example of how
 * an AppSocket connection can be established and used within
 * the context of application printer drivers, spoolers, etc.
 * Most of this code is user-interface wrapper, the basic tcp and udp methods
 * used to communicate with the printer can be discerned from the
 * source and reimplemented within other projects.
 *
 * Tektronix grants permission to use this source code, without charge,
 * in any context or project, and to modify this source for any purpose,
 * provided that this header comment is included in any derivative work 
 * in unmodified form, allowing for additions placed herein by the modifying 
 * individual or organization, provided that such additions do not alter
 * the meaning of this comment's terms.
 *
 * No warranty is provided with this source code or any executables derived
 * or compiled herefrom, nor any remedy provided for failure to perform
 * any intended function, or for any unexpected results up to and including
 * loss or data or revenue.
 *
 *
 */

/* Version string, edit to reflect updates/newer releases
 */
 
#define VERS_STR "1.2.1"

/* Defines how long the program will wait to gain a connection too
 * a printer which is reporting status: idle. Sometimes the printerr
 * will be doing internal cleanup things from a previous job, and
 * while reporting idle, will refuse new TCP connections. This value
 * controls the number of seconds between reattempts and the number
 * of reattempts before giving up (and assuming there is an authorization
 * problem.)
 */

#define IDLE_CONNECT_WAIT 5
#define IDLE_CONNECT_RETRY 60


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>

/* These constants are used to indicate the mode the program will
 * operate in as determined by command-line arguments
 */
 
typedef enum opcode
{
	DEFAULT,
	GET_STATUS,
	SEND_FILE,
	INTERACTIVE,
	EXITSERVER,
	SHOW_HELP
} opcode;

/* The cl_args structure is used to hold values and references
 * parsed from the commandline
 */
 
typedef struct cl_args
{
	int argc;						/* These are set to the exact */
	char** argv;					/* values that are passed to main() */
	char** envp;

	char* source_file_spec;		/* File (if any) to send from */

	char* target_dev_spec;		/* Network entity to attempt connection to */
	int target_dev_port;			/* Port on net target assumed to be AppSocket. */

	int quiet;						/* nonzero if don't display progress, etc. */ 
	opcode thisop;					/* Operation code (status, send file, tty */

} cl_args;

/* Global variables
 */

/* Global access to parsed commandline arguments
 */
cl_args my_args;

/* sockaddr internet structures for ourselves, the printer's status
 * and data ports.
 */
 
struct sockaddr_in my_addr, target_stat_addr, target_data_addr;

/* socket descriptors for the printer's status and data sockets
 */
 
int stat_sock_fd, data_sock_fd;

/* Global for when the child send/receive process terminates (set
 * within a signal handler, checked by the parent xmit loop.
 */

volatile int child_proc_died;

/* String containing the result of the most recent printer status
 * inquiry
 */

char last_status[256];

/* An array of characters used for pseudo-animation during transmission
 * of files to the printer via appsocket.
 */
 
char* char_ani_str = "\\|/-\\|/";

/* Text of the PostScript "executive" command.
 */

char* exec_str = "%!PS-Adobe-3.0\nexecutive\n";

/* Text of the PostScript command for exitserver with
 * executive.
 */

char* exit_exec_str = "%!PS-Adobe-3.0\n(Entering exitserver...) print flush\n"
							 "true () startjob {(successful\n)}{(failed\n)}ifelse\n"
							 "print flush\nexecutive\n";

/* Type of progress indicator to use
 */

int progress_mode;

/* constants used as indicies in the err_list string array, passed
 * to the applications error handler.
 */

enum
{
	ERR_INVALIDOP, ERR_NOFILENAME, ERR_NOPORTNUM, ERR_BADOPTFLAG,
	ERR_TARGETRESPEC, ERR_TARGETNOTSPEC, ERR_NOSTATUS, ERR_BADHOSTNAME,
	ERR_FILENOTFOUND, ERR_NOTAUTHORIZED
};

/* list of error strings for user errors. System call errors are
 * reported using perror().
 */

char* err_list[] =
{
	/* ERR_INVALIDOP */
	"Invalid option: An invalid combination of options has been specified.",
	/* ERR_NOFILENAME */
	"No file name: A file name must be specified with \"-f\" flag.",
	/* ERR_NOPORTNUM */
	"No port number: A port number must be specified with \"-p\" flag.",	
	/* ERR_BADOPTFLAG */
	"Invalid option: The specified flag is not recognized.",
	/* ERR_TARGETRESPEC */
	"Target already specified: The destination target is specified only once.",
	/* ERR_TARGETNOTSPEC */
	"Target not specified: either the PRINTER environment variable or a\n"
	"   printer host name is required.",
	/* ERR_NOSTATUS */
	"Could not get status: the printer does not respond to status requests.",
	/* ERR_BADHOSTNAME */
	"Could not read hostname: the specified hostname is an illegal format.",
	/* ERR_FILENOTFOUND */
	"Could not open the specified input file.",
	/* ERR_NOTAUTHORIZED */
	"The connection was refused from an idle printer. It is likely that\n"
	"   User authorization is enabled in the printer and this host is not\n"
	"   authorized to connect.\n"
};

/* This is the array of strings printed by the help option flag, -h.
 */
 
char* help_text[] =
{
	"The Tektronix appsocket tool may be used to perform various\n",
	"utility and communcations features to Tektronix printers with\n",
	"the TCP/IP networking option enabled. This tool uses the IP-based\n",
	"\"AppSocket\" protocols to allow bi-directional communications\n",
	"with the network printer, and to obtain status information.\n\n",
	"Usage:\n\n\tappsocket [-h | -i | -q | -f <file> | -s] [printer]\n\n",
	"Arguments enclosed in brackets \"[ ]\" are optional.\n\n",
	"Here are the option flags and their meanings:\n\n",
	"\t-h\tprint this help text and exit\n",
	"\t-i\topen an interactive session with the printer using the\n",
	"\t\tPostScript \"executive\" command. This option is\n",
	"\t\tmutually exclusive with the \"-f\" and \"-s\" options.\n",
	"\t-x\tSame as -i option, but does exitserver first.\n",
	"\t-q\tquiet operation mode. Only fatal errors are reported.\n",
	"\t\tThis mode is useful when this filter is used in scripts or\n",
	"\t\tUnix \"printcap\" files.\n",
	"\t-f <f>\tsend the file specified in the required <f>\n",
	"\t\targument to the specified printer. If \"-f\" is not specified,\n",
	"\t\tall input is taken from stdin. This option is mutually\n",
	"\t\texclusive with the \"-i\" and \"-s\" options.\n",
	"\t-s\tthis option places this tool in status mode. In status mode,\n",
	"\t\tthis tool attempts to obtain and display the PostScript status\n",
	"\t\tstring from the specified printer. The status of the printer\n",
	"\t\tis rechecked every few seconds, and if the status has changed,\n",
	"\t\tthe new status is displayed, along with a time-date stamp.\n",
	"\t\tStatus mode continues executing until a SIGTERM is received.\n",
	"\t\tThis option is mutually exclusive with both the \"-f\" and\n",
	"\t\t\"-i\" options.\n\n",
	"The \"printer\" argument may be either an Internet address in the\n",
	"form XX.XX.XX.XX, or the name of a network device which can be resolved\n",
	"by the local host. If no printer name or address is specified, the\n",
	"tool will attempt to use the name or address specified by the\n",
	"\"PRINTER\" environment variable. If no target is specified and this\n",
	"environment variable is not set, the program will terminate with an\n",
	"error.\n\n",
	""
};
	
/* Here are the function prototypes
 */

int main(int argc, char* argv[], char* envp[]);
void get_args(cl_args* a);
int get_status(char* message);
void num_to_err_str(int num);
void do_error(int err_number, char* more_info);
void resolve_target(void);
void monitor_status(int loopforever);
void reassign_stdin(void);
int process_rw(void);
char* date_time_str(void);
int send_receive_process(int src_fd, int dest_fd, int rpt_progress);
void do_help(void);

/****************************************************************************
 * main()
 *
 */

int main(int argc, char* argv[], char* envp[])
{
	int result;

	/* set the global argument and switch structure to the passed-in
	 * command arguments.
	 */
	 
	my_args.argc = argc;
	my_args.argv = argv;
	my_args.envp = envp;

	/* Process the arguments
	 */

	get_args(&my_args);

	/* Initialize the global status string to an empty string. This will
	 * force an update in the monitor_status routine the first time
	 * status is checked.
	 */
	 
	last_status[0] = '\0';
	progress_mode = 0;		/* no progress reporting is default */

	/* If the program is not in "quiet" mode, print a sign-on message.
	 */
	 
	if(!my_args.quiet || my_args.thisop == SHOW_HELP )
	{
		printf("Tektronix appsocket communication utility version %s\n",
				 VERS_STR);
		printf("Copyright (C) 1995,96 Tektronix, Inc.\n\n");
	}

	/* Now effect the option that is requested by the "thisop" global.
	 */
	 
	switch(my_args.thisop)
	{
		case GET_STATUS:

			/* Monitor status with a true (non-zero) argument causes
			 * this function to loop indefinitely.
			 */

			monitor_status(1);
			break;

		case SEND_FILE:

			/* Redirect stdin to read from the specified file in the "-f"
			 * option. Otherwise, all input is taken from the default stdin.
			 */
			 
			reassign_stdin();

			/* deliberate fallthrough to next case, since the only difference
			 * is where stdin reads from.
			 */

		case INTERACTIVE:
		case EXITSERVER:
		case DEFAULT:

			/* Process read and write. This connects to the printer and
			 * performs i/o until the stdin source is EOF, or until
			 * the printer sends a FIN and closes the connection.
			 */
			 
			result = process_rw();
			if(result != 0)
			{
				fprintf(stderr, "The connection has closed unexpectedly:\n");
				perror("result from last i/o operation");
			}
			
			break;

		case SHOW_HELP:

			/* Show the help text on the screen.
			 */
			 
			do_help();
			break;
	}

	exit(0);	
	return 0;	/* redundant, shuts up the compiler */
}

/****************************************************************************
 * get_args
 *
 * arguments: a pointer to a cl_arg structure from which the main()
 *            functions arguments may be retrieved, and into which
 *            parsed options are placed.
 *
 * returns: void, or never is an error occurs during parsing.
 *
 * get_args attempts to parse the command line switches into the
 * global status structure. Errors are reported immediately, and
 * the function does not return if errors are encountered.
 */

void get_args(cl_args* a)
{
	int t;
	char* cc;

	/* initialize the cl_args structure to default values
	 */
	 
	a->source_file_spec = NULL;	/* filename used if "-f" set */
	a->target_dev_spec = NULL;		/* name or addr. of target printer */
	a->target_dev_port = 9100;		/* default appsocket port */
	a->quiet = 0;						/* don't produce progress info if true */
	a->thisop = DEFAULT;				/* what to do, default reads stdin to */
											/* printer, reads printer to stdout */

	/* loop through & parse the arguments
	 */
	 
	for(t = 1; t < a->argc; t++)
	{

		/* set cc to a temp pointer to the current argument string.
		 */
		 
		cc = a->argv[t];

		/* if the current arg. begins with '-', treat as a switch
		 */
		 
		if(*cc == '-')
		{
			switch(*(cc + 1))
			{
				case 'q':

					/* "-q" means use quiet mode.
					 */
					 
					a->quiet = 1;
					break;

				case 'i':

					/* "-i" means go into interactive mode. First check to
					 * see if any illegal combinations (-f, -s) have been
					 * specified already.
					 */
					 
					if(a->thisop != DEFAULT)
						do_error(ERR_INVALIDOP, NULL);

					a->thisop = INTERACTIVE;
					break;

				case 'x':

					/* -x means same as -i, but exitserver first.
					 */

					if(a->thisop != DEFAULT)
						do_error(ERR_INVALIDOP, NULL);

					a->thisop = EXITSERVER;
					break;

				case 'f':

					/* "-f" option means take input from a file argument,
					 * rather than stdin. As before, check for illegal
					 * combinations of switches.
					 */

					if(a->thisop != DEFAULT)
						do_error(ERR_INVALIDOP, NULL);

					++t;
					if(t >= a->argc)
						do_error(ERR_NOFILENAME, NULL);
						
					a->source_file_spec = a->argv[t];
					a->thisop = SEND_FILE;
					break;

				case 's':

					/* User is requesting a status operation, which is
					 * mutually exclusive with -f and -i options.
					 */

					if(a->thisop == SEND_FILE)
						do_error(ERR_INVALIDOP, NULL);

					 a->thisop = GET_STATUS;
					 break;

				case 'p':

					/* User is specifying a port other than default (9100)
					 */

               ++t;
               if(t >= a->argc)
                  do_error(ERR_NOPORTNUM, NULL);

					if(sscanf(a->argv[t], "%d", &a->target_dev_port) != 1)
						do_error(ERR_NOPORTNUM, a->argv[t]);
					break;

				case 'h':

					/* User is requesting a help text string.
					 */
					 
					a->thisop = SHOW_HELP;
					break;

				default:

					/* This is an unrecognized option, print an error.
					 */
					 
					do_error(ERR_BADOPTFLAG, a->argv[t]);

			}
		} else
		{
			/* No switch, therefore consider this a target IP device
			 * specifyier (x.x.x.x or somename.somedomain...)
			 */

			if(a->target_dev_spec != NULL)
				do_error(ERR_TARGETRESPEC, a->argv[t]);

			a->target_dev_spec = a->argv[t];
			
		}
	}

	if(a->thisop != SHOW_HELP)
	{
		/* if no printer is spec'd in the command line, get one
  		 * from the environment.
		 */
		
		if(a->target_dev_spec == NULL)
		{
			a->target_dev_spec = getenv("PRINTER");
			}
		
		/* After all of these attempts, if there is STILL no printer spec'd,
		 * print an error and quit.
		 */
		
		if(a->target_dev_spec == NULL)
		{
			do_error(ERR_TARGETNOTSPEC, NULL);
			} else
		{
			/* If a printer has been specified, attempt to resolve its name,
			* and establish the socket and sockaddr_in globals.
			*/

			resolve_target();
		}	
	}

}

/****************************************************************************
 * resolve_target()
 *
 * arguments: none
 * returns: void
 *
 * resolve_target uses the global variables only to perform its function.
 * It first attempts to resolve the target host name or address into an
 * internet address (sockaddr_in and in_addr), and generates an error if
 * the target specifier is unresolvable. Next, it pre-initializes a sockaddr
 * structure for the printer's data port and another for its
 * status port, and a sockaddr for this hosts's local port.
 *
 */

void resolve_target(void)
{
	struct hostent* target_hostent;
	struct in_addr target_address;
	int b;
	
	if(sscanf(my_args.target_dev_spec, "%d.%d.%d.%d", &b, &b, &b, &b) != 4)
	{
		target_hostent = gethostbyname(my_args.target_dev_spec);
		if(target_hostent != NULL &&
			target_hostent->h_addrtype == AF_INET &&
			target_hostent->h_length == sizeof(struct in_addr))
			{
				target_address = *(struct in_addr *) target_hostent->h_addr;
			} else
			{
				do_error(ERR_BADHOSTNAME, my_args.target_dev_spec);
			}
	} else
	{
		target_address.s_addr = inet_addr(my_args.target_dev_spec);
		if((signed long) target_address.s_addr == -1)
		{
			do_error(ERR_BADHOSTNAME, my_args.target_dev_spec);
		}
	}

	/* Create internet socket addresses for our local machine, and
	 * the remote status and data ports.
	 */

	/* This is a UDP port having number of the dataport + 1. This
	 * port will echo a zero-length packet back to the sending UDP
	 * port and address, but with the msg. set to the printer's
	 * status string.
	 */
	 
	bzero((char*) &target_stat_addr, sizeof(target_stat_addr));
	target_stat_addr.sin_family = AF_INET;
	target_stat_addr.sin_addr.s_addr = target_address.s_addr;
	target_stat_addr.sin_port = htons(my_args.target_dev_port + 1);

	/* This is a TCP port having number of the dataport (default: 9100)
	 * This port will accept data from another authorized host (default:
	 * all hosts) for execution by the PS interpreter. It is a "raw" port,
	 * all 8-bit characters are valid.
	 */

	bzero((char*) &target_data_addr, sizeof(target_data_addr));
	target_data_addr.sin_family = AF_INET;
	target_data_addr.sin_addr.s_addr = target_address.s_addr;
	target_data_addr.sin_port = htons(my_args.target_dev_port);

	/* This is the local tcp or udp port.
	 */
	
	bzero((char*) &my_addr, sizeof(my_addr));
	my_addr.sin_family = AF_INET;
	my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	my_addr.sin_port = htons(0);

	/* create a socket for the status commands (local udp socket.)
	 */
	 
	if((stat_sock_fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	{
		perror("stat_sock_fd: socket failed");
		exit(1);
	}

	/* bind status socket to our local address
	 */
	 
	if(bind(stat_sock_fd, (struct sockaddr*) &my_addr, sizeof(my_addr)) < 0)
	{
		perror("stat_sock_fd: bind failed");
		exit(1);
	}


}

/****************************************************************************
 * get_status()
 *
 * arguments: a pointer to a pre-allocated character buffer which will
 *            receive the status message string.
 * returns: 0 on success, -1 if the printer could not be reached after
 *          the specified retry interval and count.
 *
 * Sends a zero-length (message field) UDP packet to the printer,
 * and it should return a UDP packet to us with its status string.
 * Since UDP packets can get lost or munged, the receive command
 * loops while the transmit packet gets resent every few moments.
 * If the remote device does not response within RETRY_COUNT
 * attempts at RETRY_INTERVAL seconds, the function returns -1,
 * otherwise (success) 0.
 */

#define STAT_RETRY_INTERVAL 1
#define STAT_RETRY_COUNT 10

int get_status(char *message)
{
	struct sockaddr_in response_addr;
	int response_len = -1, resp_sock_len;
	int attempts;
	char is_block;

	/* create a sockaddr structure to receive the response
	 * UDP packet's address. This structure's in_addr field
	 * is compared to the printer's known address, which allows
	 * us to filter a response (very, very, unlikely) from
	 * some unknown host.
	 */
	 
	bzero((char*) &response_addr, sizeof(response_addr));
	attempts = STAT_RETRY_COUNT;
	
	/* loop while the recvfrom returns failure (-1) and our
	 * retry counter has not expired.
	 */
	 
	while(response_len == -1 && attempts > 0)
	{
	
		/* set the status socket to BLOCK, so that it does
		 * NOT return until the UDP packet has been sent.
		 */
		 
		is_block = 0;
		if(ioctl(stat_sock_fd, FIONBIO, &is_block) == -1)
		{
			perror("stat_sock_fd: ioctl clear nonblock i/o failed");
			exit(1);
		}
		
		/* send the UDP "ping" packet. Note that a successful return
		 * value does NOT mean that the printer successfully received
		 * this packet.
		 */
		 
		if(sendto(stat_sock_fd, message, 0, 0,
					 (struct sockaddr*) &target_stat_addr,
					 sizeof(target_stat_addr)) < 0)
		{
			perror("stat_sock_fd: sento failed");
			exit(1);
		}

		/* now set the status socket to NON-BLOCK, so that we
		 * return immediately, whether or not there is a packet
		 * received and ready. We might wait a long, long, time
		 * if this call we issued in a blocking mode, since either
		 * our "ping" to the printer may have got lost, or the printer's
		 * response may have gone astray.
		 */
		 
		is_block = 1;
		if(ioctl(stat_sock_fd, FIONBIO, &is_block) == -1)
		{
			perror("stat_sock_fd: ioctl set nonblock i/o failed");
			exit(1);
		}

		/* Now issue a recvfrom to read the printer's response.
		 * If the recvfrom gets a good packet, but its address
		 * does not match the target printer, print a pithy little
		 * diagnostic and continue trying.
		 */
		 
		resp_sock_len = sizeof(response_addr);
		response_len = recvfrom(stat_sock_fd, message, 255, 0,
										(struct sockaddr*) &response_addr,
										&resp_sock_len);

		if(response_addr.sin_addr.s_addr !=
			target_stat_addr.sin_addr.s_addr && response_len != -1)
		{
			fprintf(stderr, "Someone else responded!!\n");
			continue;
		}

		/* now we test the result of the recvfrom to see if any
		 * data was actually received. If there was an error, it
		 * must be the EWOULDBLOCK error (meaning no received response)
		 * Other errors may indicate a system error condition, in which
		 * case we print an error and exit.
		 */
		
		if(response_len == -1)
		{	if(errno != EWOULDBLOCK)
			{
				perror("stat_sock_fd: recvfrom failed");
				exit(1);
			} else
			{
				/* There was no response packet ready, so
				 * sleep for the specified interval, decrement
				 * the retry counter, and loop to the beginning
				 * of the while() block and retry sending the "ping".
				 */
			
				sleep(STAT_RETRY_INTERVAL);
				attempts--;
			}
		}

	}
	
	/* If attempts > 0, a response was received before the counter/timer
	 * loop expired. We will add a C-string null termination character
	 * at the appropriate position in the string (indicated by response_len).
	 * also, return either 0 or -1, depending on success or failure, using
	 * the ternary operator.
	 */
	
	if(attempts > 0)
		message[response_len] = '\0';

	return (attempts > 0 ? 0 : -1);
}

/****************************************************************************
 * monitor_status()
 *
 * arguments: an integer operator, when nonzero (true) indicates this loop
 *            continue indefinitely checking the printer, and printing
 *            status changes. If false (0), print one status record and
 *            return.
 * returns: nothing
 *
 * monitor_status() is used to print the printer's status on the console,
 * along with a time/date stamp. The routine operates in two modes, depending
 * on its argument: either loop forever, printing status changes, or print
 * the status once (if changed) and return. Status changes are calculated
 * against the global last_status string, which is initialized to a zero-len
 * string, guaranteeing that status is printed the first time this function
 * is invoked. Thereafter, the status string last retrieved from the printer
 * is stored in last_status, and successive get_status() results are strcmp'd
 * against it. Only differing strings are printed, and the last_status updated
 * to the new_status when this occurs.
 */

#define STAT_UPDATE_INTERVAL 3

void monitor_status(int loop)
{
	char new_status[256];

	/* Print some progress info only in eternal loop mode and if we haven't
	 * been told to be "quiet".
	 */
	 
	if(loop && !my_args.quiet)
	{
		printf("Monitoring status for printer \"%s\"\n",
				 my_args.target_dev_spec);
		printf("Send ^C (SIGTERM) to stop.\n");
	}

	/* Here begins the eternal loop (or not) depending on whether or not
	 * we have been told to do so in the "loop" parameter. If not, status
	 * is only checked/printed/updated once.
	 */
	 
	do
	{
		/* read status from printer and check result
		 */
		 
		if(get_status(new_status) == -1)
		{
			do_error(ERR_NOSTATUS, my_args.target_dev_spec);
		}
		
		/* See if the status has changed from previously stored
		 * in "last_status" global.
		 */
		 
		if(strcmp(new_status, last_status))
		{
			/* status *has* changed, store in last_status, and print
			 * new status, along with a date/time stamp.
			 */
			 
			strcpy(last_status, new_status);
			printf("%s\t%s\n", date_time_str(), last_status);
		}

		if(loop)
			
			/* only if eternal looping is requested, sleep for a 
			 * specified number of seconds, then loop/recheck status.
			 */
			 
			sleep(STAT_UPDATE_INTERVAL);
	} while(loop);
}

/****************************************************************************
 * reassign_stdin()
 *
 * arguments: none
 * returns: nothing
 *
 * reassign_stdin is called when the -f <file> option is invoked. It attempts
 * to open via a file descriptor the file & pathname specified with the -f
 * switch. If successful, it closes the default stdin descriptor and replaces
 * it with the descriptor returned from the open file, via the dup2 call.
 */

void reassign_stdin(void)
{
	int new_input_fd;

	/* attempt to open the specified file. print a diagnostic and exit
	 * upon failure. If the specified file is "-", don't remap stdin.
	 * set progress mode to partial.
	 */

	progress_mode = 2;

	if(strcmp(my_args.source_file_spec, "-"))
	{
		if((new_input_fd = open(my_args.source_file_spec, O_RDONLY)) == -1)
		{
			do_error(ERR_FILENOTFOUND, my_args.source_file_spec);
		}

		/* remap the opened file onto stdin.
		 */
	 
		if(dup2(new_input_fd, 0) == -1)
		{
			perror("dup2 failed:");
			exit(1);
		}

		/* set the progress indicator mode to full
		 */

		progress_mode = 1;
	}
}

/****************************************************************************
 * handle_sigchld()
 *
 * arguments: none
 * returns: nothing
 *
 * this procedure is installed as a signal handler for the SIGCHLD interrupt
 * used to interrupt and flag the parent process for when the child process
 * terminates prematurely. All we do is set a global variable, which is
 * checked throughout the i/o loop process. Both parent & child i/o use the
 * same process to perform i/o, but since the child must terminate for this
 * to become true, only the parent ever sees the "true" condition and exits.
 *
 */

void handle_sigchld()
{
	child_proc_died = 1;
}

/****************************************************************************
 * process_rw()
 *
 * arguments: none
 * returns: 0 on success, -1 on failure.
 *
 * process_rw is called as the primary i/o loop initiator for all operating
 * modes except for -s "status" mode. It attempts to connect to the target
 * printer via TCP, and then forks into two processes (this parent and one
 * child) to handle the two directions of data flow. It detects when one
 * of the flows is closed, either normally (EOF) or abnormally (error), and
 * then shuts down the connection and returns. The printer-to-host flow
 * closing immediately terminates both directions of i/o, while the host-to
 * printer connection closing will allow the opposite direction connection
 * to await an EOF.
 */

#define CONN_RETRY_INTERVAL (IDLE_CONNECT_WAIT)

int process_rw(void)
{
	int connect_result = -1;
	int result, child_status;
	int pid, is_block;
	struct linger linger_struct;
	int not_idle;

	/* Print a progress message, unless told via switch to "be quiet".
	 */
	 
	if(!my_args.quiet)
	{
		printf("Attempting to connect to host printer \"%s\"...\n",
				 my_args.target_dev_spec);
	}
	
	/* start a loop while the connect() call returns error. This
	 * loop with attempt to connect forever until success, unless it seems
	 * we are not authorized. There are two conditions under which an operating 
	 * printer target will refuse connection: (1) there is a pre-existing 
	 * connection to the TCP data port of the printer, either from this host 
	 * or any other, or (2) the printer's security features have been enabled 
	 * via the PrintHost value in the AppSocket device parameter dictionary, and
	 * this host is not authorized. 
	 *
	 * If there's nobody home at the destination address, i.e. perhaps the
	 * printer is powered off, or the user specified an IP address numerically
	 * and was mistaken about that address, the monitor_status() call will
	 * indicate the nobody home error with a "can't get status" type message.
	 * and will exit this program.
	 *
	 * If there's somebody home, but the connection is refused, we'll examime
	 * the resulting status string for the "idle" keyword substring. If
	 * it's found, we are most likely facing scenario (2) above, and are not
	 * authorized. Rather than beat our head against a wall until a kindly
	 * SIGTERM puts us out of our misery, we'll abandon the attempt and
	 * exit with a diagnostic.
	 */

	not_idle = 0;
	while(connect_result == -1)
	{
		/* create a socket for tcp sending of data to appsocket
		 */

		if((data_sock_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		{
			perror("data_sock_fd: socket failed");
			exit(1);
		}

		/* bind data socket to our local address
		 */

		if(bind(data_sock_fd, (struct sockaddr*) &my_addr, sizeof(my_addr)) < 0)
		{
			perror("data_sock_fd: bind failed");
			exit(1);
		}

		/* try to open a connection via our data_sock_fd
		 */

		connect_result = connect(data_sock_fd, (void *)&target_data_addr,
								 sizeof(target_data_addr));

		/* did we get in?
		 */
		 
		if(connect_result == -1)
		{
			/* not in. call monitor_status to print the printer's status
			 * on stdout. If the printer is idle, its a good bet
			 * that we're excluded from connection due to PrintHost
			 * security. Close the socket, which will now be invalid.
			 */

			close(data_sock_fd);

			monitor_status(0);
			if(strstr(last_status, "idle") != NULL)
			{
				/* idle was found in the status string!
				 * waits until printer is idle for
				 * IDLE_CONNECT_WAIT until it decides we're
				 * not authorized.
				 */

				if(!my_args.quiet)
				{
					fprintf(stderr, ".");
				}

				if(++not_idle > IDLE_CONNECT_RETRY)
				{
					do_error(ERR_NOTAUTHORIZED, my_args.target_dev_spec);
				}
				
			} else
			{
				not_idle = 0;
			}
			
		}
		
		/* wait for the specified number of seconds, then the while()
		 * loop will retry connection.
		 */
		 
		sleep(CONN_RETRY_INTERVAL);
	}

	if(not_idle > 0 && !my_args.quiet)
	{
		fprintf(stderr, "\n\n");
	}
	
	/* Inform console of success, unless in quiet mode.
	 */
	 
	if(!my_args.quiet)
	{
		printf("...Connection established.\n\n");
	}

	/* set the socket option SO_LINGER, to send any data remaining
	 * in the buffers after our shutdown and close call.
	 */

	linger_struct.l_onoff = 1;
	linger_struct.l_linger = 1;
	if(setsockopt(data_sock_fd, SOL_SOCKET, SO_LINGER, &linger_struct,
				  sizeof(linger_struct)) < 0)
	{
		fprintf(stderr, "setsockopt->linger failed.\n");
		perror("");
		exit(1);
	}

	/* If we're in interactive mode, send the PostScript "executive"
	 * command.
	 */
	 
	if(my_args.thisop == INTERACTIVE ||
		my_args.thisop == EXITSERVER)
	{
		char* send_str;
	
		/* tell the console what we're doing, unless quiet mode is on.
		 */
		 
		if(!my_args.quiet)
		{
			if(my_args.thisop == INTERACTIVE)
			{
				send_str = exec_str;
				printf("Sending PostScript \"executive\" command...\n");
				fflush(stdout);
			} else
			{
				send_str = exit_exec_str;
				printf("Sending exitserver (startjob) and executive command...\n");
				fflush(stdout);
			}
				
		}
		
		/* write the executive string, and check for errors.
		 */
		 
		if(write(data_sock_fd, send_str, strlen(send_str)) == -1)
		{
			perror("write() error for executive string");
			exit(1);
		}
	}
	
	/* Install the SIGCHLD signal handler, and initialize the child died
	 * global boolean.
	 */
	 
	child_proc_died = 0;
	if(signal(SIGCHLD, handle_sigchld) == (void*) -1)
	{
		perror("setting SIGCHLD handler (signal) failed");
		exit(1);
	}
	
	/* now, setup two "read from/write to" processes, via a fork().
	 * one process is the parent, the other the child.
	 */

	if((pid = fork()) != 0)
	{
		/* Parent process, handles sending data from stdin to printer.
		 * pass appropriate descriptors to the send_receive_process
		 * subroutine, which handles things from here
		 */

		result = send_receive_process(0, data_sock_fd, progress_mode);
									 
		/* the result of this parent process, since it's sending
		 * from host to printer, must allow the opposite data flow
		 * channel to remain alive until the printer has finished
		 * sending any backchannel results. Accordingly, we only
		 * kill the printer-to-host flowing child process if we
		 * ourselves experienced termination of our connection direction
		 * due to an error other than normal EOF.
		 */
		 
		if(result == -1)
		{
			/* Kill the child process, since the send-to-printer
			 * reported a fatal error.
			 */

			fprintf(stderr, "Killing child process.\n");
			kill(pid, SIGKILL);
		}
		
		/* Now that we've either sent a kill to the child, or not,
		 * wait for that child to return status to us. This allows
		 * us to determine some diagnostic as well as prevent nasty
		 * zombie processes.
		 */
		 
		wait(&child_status);

	} else /* fork process result */
	{
		/* Child process, handles reading from printer to stdout.
		 * call the same send_receive_process function, but with
		 * the data flow direction reversed. Return a result when
		 * the child dies, either normally or abnormally, via exit()
		 */

		result = send_receive_process(data_sock_fd, 1, 0);
		exit(result);
		
	}
	
	/* Attempt to restore the console to normal, i.e. "blocking" i/o
	 */
	 
	is_block = 0;
	if(ioctl(0, FIONBIO, &is_block) == -1)
	{
		perror("error resetting stdin to BLOCKING mode");
	}
	
	/* Check the final exit status. 
	 */
	 
	if(result == 0 && child_status < 0)
		result = child_status;
		
	return result;
}
	
/****************************************************************************
 * send_receive_process()
 *
 * arguments: a source (read from) descriptor, and a destination (write to)
 *            descriptor, and an integer as a flag, depending on whether
 *            or not we should display progress information (host-to-printer
 *            only, in the "send file" mode of operation.
 *
 * returns:   0 on normal termination, -1 for an error.
 *
 * This subroutine is called once each from a parent process and a child
 * process, with descriptors providing for each direction of data flow; host-
 * to printer and printer-to-host. It loops through reading its source into
 * an io_buffer, and then immediately writing the same buffer back to its
 * destination. It detects EOF on the read from source, and closes the
 * loop down in that case. If the write-to-destination descriptor is a network
 * port, it calls close() to indicate EOF to the other end.
 */

#define IO_BUFSIZE 16384
#define UPD_PROG_INTERVAL 1

int send_receive_process(int src_fd, int dest_fd, int rpt_progress)
{
	unsigned char* iobuf, *write_ptr;
	int read_len, write_len;
	int at_eof, bytes_sent, src_size;
	struct stat stat_buf;
	int xmit_phase;
	char* animations;
	time_t start_time, update_time;

	/* Attempt to allocate an io buffer of IO_BUFSIZE on the heap.
	 * whine about it and exit if failure.
	 */
	 
	if((iobuf = malloc(IO_BUFSIZE)) == NULL)
	{
		perror("iobuf malloc() failed");
		return(-1);
	}

	/* If we are asked to report progress (only in "-f" mode, and
	 * only if not asked to be quiet, we will be printing progress
	 * in the file send to stderr. For our statistics, we'll need
	 * the size of the file to be sent, obtained via fstat().
	 */
	 
	if(rpt_progress)
	{
		if(fstat(src_fd, &stat_buf) == -1)
		{
			perror("fstat() failed on source descriptor");
			return(-1);
		}

		src_size = stat_buf.st_size;
	}

	/* Set the system to non-restarting system calls (signal interrupt
	 * behavior
	 */

	siginterrupt(SIGCHLD, 1);

	/* Initialize our eof marker and also some of the statistics
	 * counters.
	 */
	 
	bytes_sent = 0;
	xmit_phase = 1;
	at_eof = 1;
	start_time = time(NULL);
	update_time = 0;

	/* this while() begins the read/write loop. It does not exit
	 * this loop until an EOF was read from the source descriptor, or
	 * until the child_died global flag is set (parent version only
	 * will ever see this true.)
	 */
	 
	while(at_eof && !child_proc_died)
	{
		/* read as many bytes as possible from the source descriptor
		 * up to IO_BUFSIZE. a zero length read will indicate EOF. This
		 * descriptor is set to non-blocking (above) so that the child_proc_died
		 * flagging will work correctly.
		 *
		 */
		
		read_len = -1;
		while(!child_proc_died && read_len < 0)
		{
			if((read_len = read(src_fd, iobuf, IO_BUFSIZE)) == -1)
			{

				/* read failed - check for a signal (child died?),
				 * and reissue the read call if so. This will force
				 * a check of child_proc_died.
				 */
				 
				if(errno != EINTR)
				{
					perror("read() error");
					return(-1);
				}
			}
		}

		/* is eof?
		 */

		if(read_len == 0)
		{
			at_eof = 0;
		
		} else
		{
			/* not eof. write the recently read buffer out to the
			 * destination descriptor. Since network writes may not
			 * write all requested data in any single call, we must
			 * track how many bytes were written, and requeue the write
			 * until all the data have been sent. We advance a pointer
			 * to keep track of the position in the buffer in the instance
			 * of multiple writes required.
			 */
			
			write_ptr = iobuf;

			while(!child_proc_died && read_len > 0)
			{
				write_len = write(dest_fd, write_ptr, read_len);
				write_ptr += write_len;
				
				if(write_len == -1)
				{
					/* If this is an interrupted write call, re-loop,
					 * which will test for eof & child process died
					 * again
					 */
					 
					if(errno == EINTR)
					{
						continue;

					} else
					{
					
						/* write failed - return with an error to our
						 * caller.
						 */
					 
						perror("write() error");
						return(-1);
					}
				}
				
				/* update our progress statistics, and see if there
				 * are more data to be sent in the buffer.
				 */

				bytes_sent += write_len;
				read_len -= write_len;
			}
			
			/* now, if we've been requested to keep progress info (-f mode
			 * on the host-to-printer flow direction only, and if we've
			 * not been told to keep quiet, we'll update our progress indicators
			 * every bufferful in the read/write loop.
			 */
			 
			if(rpt_progress && !my_args.quiet)
			{
				if(time(NULL) - update_time > UPD_PROG_INTERVAL)
				{

					/* the "phase" variable is used to print animated character
					* graphics: "-" "\" "|" "/" twice through, then an update
					* to the statistics string. There are four phases, and we
					* loop through the phase loop twice, but the initial marker
					* "-" is printed as part of the statistics string, so
					* the character array is seven elements long.
					*/
				 
					if(--xmit_phase > 0)
					{
						/* we're still in the dancing character mode. Print
						* a CR *without* an LF and draw the new character.
						*/
					 
						putc('\r', stderr);
						putc(' ', stderr);
						putc(' ', stderr);
						putc(*animations++, stderr);
					
					} else
					{
						/* print the statistics line with a CR, NOT an LF, and reset
						* the phase counter and character string pointer.
						*/

						if(rpt_progress == 1)
						{
							/* input is disk-based, so we can get the length and
							* therefore report completion status.
							*/
					
							fprintf(stderr,
									  "\r  - * Sending file, %d bytes of %d; %.02f%% complete",
									  bytes_sent, src_size,
									  (double) bytes_sent / (double) src_size * 100.0);
						} else
						{
							/* input is stream based, therefore EOF length is
							 * not known and only to-date data sent can be
							 * reported.
							 */
						 
							fprintf(stderr,
									  "\r  - * Sending file, %d bytes", bytes_sent);
						}
					
						xmit_phase = 8;
						animations = char_ani_str;
					}

					update_time = time(NULL);
					
				} /* end of show update time interval expired statement */
					
			} /* end of report on progress if statement */
			
		} /* end "haven't read eof, so do write" if statement */
	} /* end of read/write while not eof && !child_proc_died loop */

	/* Restore the signal system call restarting status to normal
	 */

	siginterrupt(SIGCHLD, 0);

	/* de-allocate the IO buffer.
	 */

	free(iobuf);

	/* since we got here, we've read from our input descriptor until
	 * eof, now check to see if we were writing to the network, in
	 * which case we xmit a close, to indicate EOF to the printer.
	 */

	if(dest_fd > 2)
	{
		shutdown(dest_fd, 1);	/* no more to send */
	}

	if(close(dest_fd) == -1)
	{
		perror("close() failed");
		return(-1);
	}

	if(rpt_progress && !my_args.quiet)
	{
		int hh, mm, ss, elapsed;
		double throughput;
		elapsed = time(NULL) - start_time;

		if(elapsed > 0)
		{
			throughput = (double) bytes_sent / (double) elapsed;
		} else
		{
			throughput = bytes_sent;
		}
		
		throughput /= 1024.0;	/* to KB/sec */
		
		hh = elapsed / (60 * 60);
		elapsed %= (60 * 60);
		mm = elapsed / 60;
		elapsed %= 60;
		ss = elapsed;
			
		fprintf(stderr, "\rTotal bytes sent: %d in %02dh:%02dm:%02ds (%.2f K/sec)\n",
				  bytes_sent, hh, mm, ss, throughput);
	}

	/* return no error.
	 */
	 
	return(0);
}

/****************************************************************************
 * do_help()
 *
 * arguments: none
 * returns: never returns
 *
 * do_help prints all of the elements in the help_text string array.
 * the end of the array is indicated by a zero length ( "" ) string.
 */

void do_help(void)
{
	int str_index;

	str_index = 0;

	while(*help_text[str_index] != '\0')
		fputs(help_text[str_index++], stderr);
		
	exit(1);
}

/****************************************************************************
 * date_time_str()
 *
 * arguments: none
 * returns: a pointer to a string containing the current system date and
 *          local time.
 *
 * This function is called to provide date & time stamps for whatever reason.
 * the char* string is globally allocated by the standard libraries. (see
 * man page entry for ctime() )
 */

char* date_time_str(void)
{
   struct timeval tp;

   gettimeofday(&tp, NULL);

   return ctime(&(tp.tv_sec));
}

/****************************************************************************
 * do_error()
 *
 * arguments: an integer (actually an enum) which serves as a selector
 *            into the globally defined error string array, and a string
 *            which can provide additional information, such as an offending
 *            filename or option switch. This pointer should be NULL if no
 *            additional information is provided.
 *
 * returns: never returns (exit's)
 *
 * do_error is called whenever a non-system, user caused, error is encountered, 
 * such as invalid argument's parsed, bad printer destination address, 
 * bad filename, etc. System errors which indicate abnormal, unexpected program 
 * execution due to non-user related factors are handled on the spot when 
 * detected by the perror() system call and an immediate exit().
 */

void do_error(int err_number, char* moreinfo)
{
	/* print the error string indexed by the error number to stderr.
	 */
	 
	fprintf(stderr, "ERROR: %s\n", err_list[err_number]);
	
	/* is there an additional error info string as evidenced by a non-NULL
	 * moreinfo pointer?
	 */
	 
	if(moreinfo != NULL)
		fprintf(stderr, "Offending argument: %s\n", moreinfo);

	/* print a helpful message to steer the pilot to the help screen.
	 */
	 
	fprintf(stderr, "Use \"%s -h\" for help.\n", my_args.argv[0]);
	
	/* eject, Buckaroo, eject!
	 */
	 
	exit(1);
}

/* end-of-program source file: appsocket.c
 */
